import 'dart:math';

import 'package:equatable/equatable.dart';
import 'package:flutter/material.dart';
import 'package:hydrated_bloc/hydrated_bloc.dart';

import 'package:tetrisdual/models/game/game.dart';
import 'package:tetrisdual/models/game/player.dart';
import 'package:tetrisdual/models/settings/settings_game.dart';
import 'package:tetrisdual/models/settings/settings_global.dart';

part 'game_state.dart';

class GameCubit extends HydratedCubit<GameState> {
  GameCubit()
      : super(GameState(
          currentGame: Game.createEmpty(),
        ));

  void updateState(Game game) {
    emit(GameState(
      currentGame: game,
    ));
  }

  void refresh() {
    final Game game = Game(
      // Settings
      gameSettings: state.currentGame.gameSettings,
      globalSettings: state.currentGame.globalSettings,
      // State
      isRunning: state.currentGame.isRunning,
      isStarted: state.currentGame.isStarted,
      isFinished: state.currentGame.isFinished,
      animationInProgress: state.currentGame.animationInProgress,
      // Base data
      players: state.currentGame.players,
      // Game data
      currentPlayer: state.currentGame.currentPlayer,
    );
    // game.dump();

    updateState(game);
  }

  void startNewGame({
    required GameSettings gameSettings,
    required GlobalSettings globalSettings,
  }) {
    final Game newGame = Game.createNew(
      // Settings
      gameSettings: gameSettings,
      globalSettings: globalSettings,
    );

    newGame.dump();

    updateState(newGame);
    enableRandomPlayer();
    refresh();
  }

  void quitGame() {
    state.currentGame.isRunning = false;
    refresh();
  }

  void resumeSavedGame() {
    state.currentGame.isRunning = true;
    refresh();
  }

  void deleteSavedGame() {
    state.currentGame.isRunning = false;
    state.currentGame.isFinished = true;
    refresh();
  }

  void toggleCurrentPlayer() {
    state.currentGame.isStarted = true;

    // brand new game
    if (state.currentGame.currentPlayer == 0) {
      state.currentGame.currentPlayer = 1;
    } else {
      // Reset current player tetrimino
      getCurrentPlayer().resetTetrimino();

      // toggle: 1 -> 2 ; 2 -> 1
      state.currentGame.currentPlayer = 3 - state.currentGame.currentPlayer;
    }

    // Pick new tetrimino
    getCurrentPlayer().pickRandomTetrimino();

    refresh();
  }

  void enableRandomPlayer() {
    state.currentGame.currentPlayer = Random().nextInt(2) + 1;
    toggleCurrentPlayer();
  }

  Player getPlayer(int playerId) {
    return state.currentGame.players[playerId - 1];
  }

  Player getCurrentPlayer() {
    return getPlayer(state.currentGame.currentPlayer);
  }

  @override
  GameState? fromJson(Map<String, dynamic> json) {
    final Game currentGame = json['currentGame'] as Game;

    return GameState(
      currentGame: currentGame,
    );
  }

  @override
  Map<String, dynamic>? toJson(GameState state) {
    return <String, dynamic>{
      'currentGame': state.currentGame.toJson(),
    };
  }
}
